from pymatgen.util.testing import PymatgenTest
from pymatgen.core import Structure

struct_names = (
    "SiO2 Li2O LiFePO4 TlBiSe2 K2O2 Li3V2(PO4)3 Li2O2 CsCl NaFePO4 Pb2TiZrO6 "
    "SrTiO3 TiO2 BaNiO3 VO2".split()
)

test_structures = {
    name.stem: Structure.from_file(name)
    for name in PymatgenTest.TEST_STRUCTURES.keys()
    if name.stem in struct_names
}



from typing import TYPE_CHECKING, Dict, List, Optional, Union

from matcalc.elasticity import ElasticityCalc
from matcalc.eos import EOSCalc
from matcalc.phonon import PhononCalc
from matcalc.relaxation import RelaxCalc
from matcalc.util import get_universal_calculator
from pydantic import Field, validator
from pymatgen.analysis.elasticity import ElasticTensor
from pymatgen.core import Structure

from emmet.core.elasticity import (
    BulkModulus,
    ElasticityDoc,
    ElasticTensorDoc,
    ShearModulus,
)

if TYPE_CHECKING:
    from ase.calculators.calculator import Calculator


class MLDoc(ElasticityDoc):
    """Document model for matcalc-generated material properties from machine learning
    interatomic potential predictions.

    Attributes:
    - metadata
        - material_id (str): MP ID
        - structure (Structure): pymatgen Structure object
        - deprecated (bool): whether this material is deprecated in MP
        - calculator (str): name of model used as ML potential.
        - version (str): version of matcalc used to generate this document
    - relaxation
        - final_structure: relaxed pymatgen Structure object
        - energy (float): final energy in eV
        - volume (float): final volume in Angstrom^3
        - lattice parameters (float): a, b, c, alpha, beta, gamma
    - equation of state
        - eos (dict[str, list[float]]): with keys energies and volumes
        - bulk_modulus_bm (float): Birch-Murnaghan bulk modulus in GPa
    - phonon
        - temperatures (list[float]): temperatures in K
        - free_energy (list[float]): Helmholtz energies at those temperatures in eV
        - entropy (list[float]): entropies at those temperatures in eV/K
        - heat_capacities (list[float]): heat capacities at constant volume in eV/K
    - elasticity
        - elastic_tensor (ElasticTensorDoc): pydantic model from emmet.core.elasticity
        - shear_modulus (ShearModulus): Voigt-Reuss-Hill shear modulus (single float in eV/Angstrom^3)
        - bulk_modulus (BulkModulus): Voigt-Reuss-Hill bulk modulus (single float in eV/Angstrom^3)
        - youngs_modulus (float): Young's modulus (single float in eV/Angstrom^3)
    """

    property_name: str = "ml"

    # metadata
    structure: Structure = Field(description="Original structure")
    matcalc_version: Optional[str] = Field(
        None, description="Version of matcalc used to generate this document"
    )
    model_name: Optional[str] = Field(
        None, description="Name of model used as ML potential."
    )
    model_version: Optional[str] = Field(
        None, description="Version of model used as ML potential"
    )

    # relaxation attributes
    final_structure: Optional[Structure] = Field(
        None, description="ML-potential-relaxed structure"
    )
    energy: Optional[float] = Field(None, description="Final energy in eV")
    volume: Optional[float] = Field(None, description="Final volume in Angstrom^3")
    a: Optional[float] = Field(None, description="Lattice length a in Angstrom")
    b: Optional[float] = Field(None, description="Lattice length b in Angstrom")
    c: Optional[float] = Field(None, description="Lattice length c in Angstrom")
    alpha: Optional[float] = Field(None, description="Lattice angle alpha in degrees")
    beta: Optional[float] = Field(None, description="Lattice angle beta in degrees")
    gamma: Optional[float] = Field(None, description="Lattice angle gamma in degrees")

    # equation of state attributes
    eos: Optional[Dict[str, List[float]]] = Field(
        None, description="dict with keys energies and volumes"
    )
    bulk_modulus_bm: Optional[float] = Field(None, description="bm.b0_GPa")

    # phonons attributes
    temperatures: Optional[List[float]] = Field(
        None, description="list of temperatures"
    )
    free_energy: Optional[List[float]] = Field(
        None,
        description="list of Helmholtz free energies at corresponding temperatures",
    )
    entropy: Optional[List[float]] = Field(
        None, description="list of entropies at corresponding temperatures in eV/K"
    )
    heat_capacity: Optional[List[float]] = Field(
        None,
        description="list of heat capacities at constant volume at corresponding "
        "temperatures in eV/K",
    )

    # elasticity attributes
    # all inherited from ElasticityDoc

    @validator("elastic_tensor", pre=True)
    def elastic_tensor(cls, val) -> ElasticTensorDoc:
        """ElasticTensorDoc from MSONable dict of ElasticTensor, or list (specifying the Voigt array)
        or the ElasticTensor class itself.
        """
        if isinstance(val, dict):
            tensor = ElasticTensor.from_dict(val)
        elif isinstance(val, (list, tuple)):
            tensor = ElasticTensor(val)
        else:
            tensor = val
        return ElasticTensorDoc(raw=tensor.voigt.tolist())

    @validator("bulk_modulus", pre=True, always=True)
    def bulk_vrh_no_suffix(cls, new_key, values):
        """Map field bulk_modulus_vrh to bulk_modulus."""
        val = values.get("bulk_modulus_vrh", new_key)
        return BulkModulus(vrh=val)

    @validator("shear_modulus", pre=True, always=True)
    def shear_vrh_no_suffix(cls, new_key, values):
        """Map field shear_modulus_vrh to shear_modulus."""
        val = values.get("shear_modulus_vrh", new_key)
        return ShearModulus(vrh=val)

    def __init__(
        cls,
        structure,
        material_id,
        calculator: Union[str, "Calculator"],
        prop_kwargs: Optional[dict] = None,
        **kwargs,
    ) -> None:
        """
        Args:
            structure (Structure): Pymatgen Structure object.
            material_id (str): MP ID.
            calculator (str | Calculator): ASE calculator or name of model to use as ML
                potential. See matcalc.util.UNIVERSAL_CALCULATORS for recognized names.
            prop_kwargs (dict): Keyword arguments for each matcalc PropCalc class.
                Recognized keys are RelaxCalc, ElasticityCalc, PhononCalc, EOSCalc.
            **kwargs: Passed to the PropertyDoc constructor.

        Returns:
            MLDoc
        """
        calculator = get_universal_calculator(calculator)

        results = {}
        for prop_cls in (RelaxCalc, PhononCalc, EOSCalc, ElasticityCalc):
            kwds = (prop_kwargs or {}).get(prop_cls.__name__, {})
            output = prop_cls(calculator, **kwds).calc(structure)
            # extract thermal_properties from PhononCalc output
            if isinstance(output, dict) and {*output} == {
                "phonon",
                "thermal_properties",
            }:
                output = output["thermal_properties"]
            results.update(output)

        for key, val in results.items():
            # convert arrays to lists
            if hasattr(val, "tolist"):
                results[key] = val.tolist()

        super().__init__(
            structure=structure, material_id=material_id, **results, **kwargs
        )



""" Core definition of a Materials Document """
from typing import Dict, List, Mapping, Optional
from emmet.core.base import EmmetMeta

from pydantic import Field, BaseModel
from pymatgen.analysis.structure_analyzer import SpacegroupAnalyzer
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.entries.computed_entries import ComputedStructureEntry

from emmet.core.material import MaterialsDoc as CoreMaterialsDoc
from emmet.core.material import PropertyOrigin
from emmet.core.settings import EmmetSettings
from emmet.core.structure import StructureMetadata
from emmet.core.vasp.calc_types import CalcType, RunType, TaskType
from emmet.core.vasp.task_valid import TaskDocument

SETTINGS = EmmetSettings()


class BlessedCalcs(BaseModel):
    GGA: Optional[ComputedStructureEntry] = None
    GGA_U: Optional[ComputedStructureEntry] = Field(None, alias="GGA+U")
    PBESol: Optional[ComputedStructureEntry] = None
    SCAN: Optional[ComputedStructureEntry] = None
    R2SCAN: Optional[ComputedStructureEntry] = None
    HSE: Optional[ComputedStructureEntry] = None


class MaterialsDoc(CoreMaterialsDoc, StructureMetadata):
    calc_types: Mapping[str, CalcType] = Field(  # type: ignore
        None,
        description="Calculation types for all the calculations that make up this material",
    )
    task_types: Optional[Mapping[str, TaskType]] = Field(
        None,
        description="Task types for all the calculations that make up this material",
    )
    run_types: Optional[Mapping[str, RunType]] = Field(
        None,
        description="Run types for all the calculations that make up this material",
    )

    origins: Optional[List[PropertyOrigin]] = Field(
        None, description="Mappingionary for tracking the provenance of properties"
    )

    entries: Optional[BlessedCalcs] = Field(
        None, description="Dictionary for tracking entries for VASP calculations"
    )

    @classmethod
    def from_tasks(
        cls,
        task_group: List[TaskDocument],
        structure_quality_scores: Dict[
            str, int
        ] = SETTINGS.VASP_STRUCTURE_QUALITY_SCORES,
        use_statics: bool = SETTINGS.VASP_USE_STATICS,
        commercial_license: bool = True,
    ) -> "MaterialsDoc":
        """
        Converts a group of tasks into one material

        Args:
            task_group: List of task document
            structure_quality_scores: quality scores for various calculation types
            use_statics: Use statics to define a material
            commercial_license: Whether the data should be licensed with BY-C (otherwise BY-NC).
        """
        if len(task_group) == 0:
            raise Exception("Must have more than one task in the group.")

        # Metadata
        last_updated = max(task.last_updated for task in task_group)
        created_at = min(task.completed_at for task in task_group)
        task_ids = list({task.task_id for task in task_group})

        deprecated_tasks = {task.task_id for task in task_group if not task.is_valid}
        run_types = {task.task_id: task.run_type for task in task_group}
        task_types = {task.task_id: task.task_type for task in task_group}
        calc_types = {task.task_id: task.calc_type for task in task_group}

        structure_optimizations = [
            task for task in task_group if task.task_type == TaskType.Structure_Optimization  # type: ignore
        ]
        statics = [task for task in task_group if task.task_type == TaskType.Static]  # type: ignore
        structure_calcs = (
            structure_optimizations + statics
            if use_statics
            else structure_optimizations
        )

        # Material ID
        possible_mat_ids = [task.task_id for task in structure_optimizations]

        if use_statics:
            possible_mat_ids += [task.task_id for task in statics]

        material_id = min(possible_mat_ids)

        # Always prefer a static over a structure opt
        structure_task_quality_scores = {"Structure Optimization": 1, "Static": 2}

        def _structure_eval(task: TaskDocument):
            """
            Helper function to order structures optimization and statics calcs by
            - Functional Type
            - Spin polarization
            - Special Tags
            - Energy
            """

            task_run_type = task.run_type
            _SPECIAL_TAGS = ["LASPH", "ISPIN"]
            special_tags = sum(
                task.input.parameters.get(tag, False) for tag in _SPECIAL_TAGS
            )

            return (
                -1 * int(task.is_valid),
                -1 * structure_quality_scores.get(task_run_type.value, 0),
                -1 * structure_task_quality_scores.get(task.task_type.value, 0),
                -1 * special_tags,
                task.output.energy_per_atom,
            )

        best_structure_calc = sorted(structure_calcs, key=_structure_eval)[0]
        structure = best_structure_calc.output.structure

        # Initial Structures
        initial_structures = [task.input.structure for task in task_group]
        sm = StructureMatcher(
            ltol=0.1, stol=0.1, angle_tol=0.1, scale=False, attempt_supercell=False
        )
        initial_structures = [
            group[0] for group in sm.group_structures(initial_structures)
        ]

        # Deprecated
        deprecated = all(task.task_id in deprecated_tasks for task in structure_calcs)
        deprecated = deprecated or best_structure_calc.task_id in deprecated_tasks

        # Origins
        origins = [
            PropertyOrigin(
                name="structure",
                task_id=best_structure_calc.task_id,
                last_updated=best_structure_calc.last_updated,
            )
        ]

        # Entries
        # **current materials docs must contain at last one GGA or GGA+U entry

        # Always prefer a static over a structure opt
        entry_task_quality_scores = {"Structure Optimization": 1, "Static": 2}

        def _entry_eval(task: TaskDocument):
            """
            Helper function to order entries and statics calcs by
            - Spin polarization
            - Special Tags
            - Energy
            """

            _SPECIAL_TAGS = ["LASPH", "ISPIN"]
            special_tags = sum(
                task.input.parameters.get(tag, False) for tag in _SPECIAL_TAGS
            )

            return (
                -1 * int(task.is_valid),
                -1 * entry_task_quality_scores.get(task.task_type.value, 0),
                -1 * special_tags,
                task.output.energy_per_atom,
            )

        # Entries
        # **current materials docs must contain at last one GGA or GGA+U entry
        entries = {}
        all_run_types = set(run_types.values())

        for rt in all_run_types:
            relevant_calcs = sorted(
                [doc for doc in structure_calcs if doc.run_type == rt and doc.is_valid],
                key=_entry_eval,
            )

            if len(relevant_calcs) > 0:
                best_task_doc = relevant_calcs[0]
                entry = best_task_doc.structure_entry
                entry.data["task_id"] = entry.entry_id
                entry.data["material_id"] = material_id
                entry.entry_id = "{}-{}".format(material_id, rt.value)
                entries[rt] = entry

        if RunType.GGA not in entries and RunType.GGA_U not in entries:
            raise ValueError(
                "Individual material entry must contain at least one GGA or GGA+U calculation"
            )

        # Builder meta and license
        builder_meta = EmmetMeta(license="BY-C" if commercial_license else "BY-NC")

        return cls.from_structure(
            structure=structure,
            material_id=material_id,
            last_updated=last_updated,
            created_at=created_at,
            task_ids=task_ids,
            calc_types=calc_types,
            run_types=run_types,
            task_types=task_types,
            initial_structures=initial_structures,
            deprecated=deprecated,
            deprecated_tasks=deprecated_tasks,
            origins=origins,
            entries=entries,
            builder_meta=builder_meta,
        )

    @classmethod
    def construct_deprecated_material(
        cls,
        task_group: List[TaskDocument],
        commercial_license: bool = True,
    ) -> "MaterialsDoc":
        """
        Converts a group of tasks into a deprecated material

        Args:
            task_group: List of task document
            commercial_license: Whether the data should be licensed with BY-C (otherwise BY-NC).
        """
        if len(task_group) == 0:
            raise Exception("Must have more than one task in the group.")

        # Metadata
        last_updated = max(task.last_updated for task in task_group)
        created_at = min(task.completed_at for task in task_group)
        task_ids = list({task.task_id for task in task_group})

        deprecated_tasks = {task.task_id for task in task_group}
        run_types = {task.task_id: task.run_type for task in task_group}
        task_types = {task.task_id: task.task_type for task in task_group}
        calc_types = {task.task_id: task.calc_type for task in task_group}

        # Material ID
        material_id = min([task.task_id for task in task_group])

        # Choose any random structure for metadata
        structure = SpacegroupAnalyzer(
            task_group[0].output.structure, symprec=0.1
        ).get_conventional_standard_structure()

        # Deprecated
        deprecated = True

        # Builder meta and license
        builder_meta = EmmetMeta(license="BY-C" if commercial_license else "BY-NC")

        return cls.from_structure(
            structure=structure,
            material_id=material_id,
            last_updated=last_updated,
            created_at=created_at,
            task_ids=task_ids,
            calc_types=calc_types,
            run_types=run_types,
            task_types=task_types,
            deprecated=deprecated,
            deprecated_tasks=deprecated_tasks,
            builder_meta=builder_meta,
        )



""" Core definition of a Thermo Document """
from collections import defaultdict
from typing import Dict, List, Optional, Union
from datetime import datetime
from emmet.core.base import EmmetMeta
from emmet.core.utils import ValueEnum

from pydantic import BaseModel, Field
from pymatgen.analysis.phase_diagram import PhaseDiagram
from pymatgen.entries.computed_entries import ComputedEntry, ComputedStructureEntry

from emmet.core.material_property import PropertyDoc
from emmet.core.material import PropertyOrigin
from emmet.core.mpid import MPID
from emmet.core.vasp.calc_types.enums import RunType


class DecompositionProduct(BaseModel):
    """
    Entry metadata for a decomposition process
    """

    material_id: Optional[MPID] = Field(
        None,
        description="The Materials Project ID for the material this decomposition points to.",
    )
    formula: Optional[str] = Field(
        None,
        description="The formula of the decomposed material this material decomposes to.",
    )
    amount: Optional[float] = Field(
        None,
        description="The amount of the decomposed material by formula units this this material decomposes to.",
    )


class ThermoType(ValueEnum):
    GGA_GGA_U = "GGA_GGA+U"
    GGA_GGA_U_R2SCAN = "GGA_GGA+U_R2SCAN"
    R2SCAN = "R2SCAN"
    UNKNOWN = "UNKNOWN"


class ThermoDoc(PropertyDoc):
    """
    A thermo entry document
    """

    property_name: str = "thermo"

    thermo_type: Union[ThermoType, RunType] = Field(
        ...,
        description="Functional types of calculations involved in the energy mixing scheme.",
    )

    thermo_id: str = Field(
        ...,
        description="Unique document ID which is composed of the Material ID and thermo data type.",
    )

    uncorrected_energy_per_atom: float = Field(
        ..., description="The total DFT energy of this material per atom in eV/atom."
    )

    energy_per_atom: float = Field(
        ...,
        description="The total corrected DFT energy of this material per atom in eV/atom.",
    )

    energy_uncertainy_per_atom: Optional[float] = Field(None, description="")

    formation_energy_per_atom: Optional[float] = Field(
        None, description="The formation energy per atom in eV/atom."
    )

    energy_above_hull: float = Field(
        ..., description="The energy above the hull in eV/Atom."
    )

    is_stable: bool = Field(
        False,
        description="Flag for whether this material is on the hull and therefore stable.",
    )

    equilibrium_reaction_energy_per_atom: Optional[float] = Field(
        None,
        description="The reaction energy of a stable entry from the neighboring equilibrium stable materials in eV."
        " Also known as the inverse distance to hull.",
    )

    decomposes_to: Optional[List[DecompositionProduct]] = Field(
        None,
        description="List of decomposition data for this material. Only valid for metastable or unstable material.",
    )

    decomposition_enthalpy: Optional[float] = Field(
        None,
        description="Decomposition enthalpy as defined by `get_decomp_and_phase_separation_energy` in pymatgen.",
    )

    decomposition_enthalpy_decomposes_to: Optional[List[DecompositionProduct]] = Field(
        None,
        description="List of decomposition data associated with the decomposition_enthalpy quantity.",
    )

    energy_type: str = Field(
        ...,
        description="The type of calculation this energy evaluation comes from.",
    )

    entry_types: List[str] = Field(
        description="List of available energy types computed for this material."
    )

    entries: Dict[str, Union[ComputedEntry, ComputedStructureEntry]] = Field(
        ...,
        description="List of all entries that are valid for this material."
        " The keys for this dictionary are names of various calculation types.",
    )

    @classmethod
    def from_entries(
        cls,
        entries: List[Union[ComputedEntry, ComputedStructureEntry]],
        thermo_type: Union[ThermoType, RunType],
        phase_diagram: Optional[PhaseDiagram] = None,
        use_max_chemsys: bool = False,
        **kwargs
    ):
        """Produce a list of ThermoDocs from a list of Entry objects

        Args:
            entries (List[Union[ComputedEntry, ComputedStructureEntry]]): List of Entry objects
            thermo_type (Union[ThermoType, RunType]): Thermo type
            phase_diagram (Optional[PhaseDiagram], optional): Already built phase diagram. Defaults to None.
            use_max_chemsys (bool, optional): Whether to only produce thermo docs for materials
                that match the largest chemsys represented in the list. Defaults to False.

        Returns:
            List[ThermoDoc]: List of built thermo doc objects.
        """

        pd = phase_diagram or cls.construct_phase_diagram(entries)

        chemsys = "-".join(sorted([str(e) for e in pd.elements]))

        docs = []

        entries_by_mpid = defaultdict(list)
        for e in entries:
            entries_by_mpid[e.data["material_id"]].append(e)

        entry_quality_scores = {"GGA": 1, "GGA+U": 2, "SCAN": 3, "R2SCAN": 4}

        def _energy_eval(entry: ComputedStructureEntry):
            """
            Helper function to order entries for thermo energy data selection
            - Run type
            - LASPH
            - Energy
            """

            return (
                -1 * entry_quality_scores.get(entry.data["run_type"], 0),
                -1 * int(entry.data.get("aspherical", False)),
                entry.energy,
            )

        for material_id, entry_group in entries_by_mpid.items():
            if (
                use_max_chemsys
                and entry_group[0].composition.chemical_system != chemsys
            ):
                continue

            sorted_entries = sorted(entry_group, key=_energy_eval)

            blessed_entry = sorted_entries[0]

            (decomp, ehull) = pd.get_decomp_and_e_above_hull(blessed_entry)

            builder_meta = EmmetMeta(license=blessed_entry.data.get("license"))

            d = {
                "thermo_id": "{}_{}".format(material_id, str(thermo_type)),
                "material_id": material_id,
                "thermo_type": thermo_type,
                "uncorrected_energy_per_atom": blessed_entry.uncorrected_energy
                / blessed_entry.composition.num_atoms,
                "energy_per_atom": blessed_entry.energy
                / blessed_entry.composition.num_atoms,
                "formation_energy_per_atom": pd.get_form_energy_per_atom(blessed_entry),
                "energy_above_hull": ehull,
                "is_stable": blessed_entry in pd.stable_entries,
                "builder_meta": builder_meta.model_dump(),
            }

            # Uncomment to make last_updated line up with materials.
            # if "last_updated" in blessed_entry.data:
            #     d["last_updated"] = blessed_entry.data["last_updated"]

            # Store different info if stable vs decomposes
            if d["is_stable"]:
                d[
                    "equilibrium_reaction_energy_per_atom"
                ] = pd.get_equilibrium_reaction_energy(blessed_entry)
            else:
                d["decomposes_to"] = [
                    {
                        "material_id": de.data["material_id"],
                        "formula": de.composition.formula,
                        "amount": amt,
                    }
                    for de, amt in decomp.items()
                ]

            try:
                decomp, energy = pd.get_decomp_and_phase_separation_energy(
                    blessed_entry
                )
                d["decomposition_enthalpy"] = energy
                d["decomposition_enthalpy_decomposes_to"] = [
                    {
                        "material_id": de.data["material_id"],
                        "formula": de.composition.formula,
                        "amount": amt,
                    }
                    for de, amt in decomp.items()
                ]
            except ValueError:
                # try/except so this quantity does not take down the builder if it fails:
                # it includes an optimization step that can be fragile in some instances,
                # most likely failure is ValueError, "invalid value encountered in true_divide"
                d["warnings"] = [
                    "Could not calculate decomposition enthalpy for this entry."
                ]

            d["energy_type"] = blessed_entry.parameters.get("run_type", "Unknown")
            d["entry_types"] = []
            d["entries"] = {}

            # Currently, each entry group contains a single entry due to how the compatibility scheme works
            for entry in entry_group:
                d["entry_types"].append(entry.parameters.get("run_type", "Unknown"))
                d["entries"][entry.parameters.get("run_type", "Unknown")] = entry

            d["origins"] = [
                PropertyOrigin(
                    name="energy",
                    task_id=blessed_entry.data["task_id"],
                    last_updated=d.get("last_updated", datetime.utcnow()),
                )
            ]

            docs.append(
                ThermoDoc.from_structure(
                    meta_structure=blessed_entry.structure, **d, **kwargs
                )
            )

        return docs

    @staticmethod
    def construct_phase_diagram(entries) -> PhaseDiagram:
        """
        Efficienty construct a phase diagram using only the lowest entries at every composition
        represented in the entry data passed.

        Args:
            entries (List[ComputedStructureEntry]): List of corrected pymatgen entry objects.

        Returns:
            PhaseDiagram: Pymatgen PhaseDiagram object
        """
        entries_by_comp = defaultdict(list)
        for e in entries:
            entries_by_comp[e.composition.reduced_formula].append(e)

        # Only use lowest entry per composition to speed up QHull in Phase Diagram
        reduced_entries = [
            sorted(comp_entries, key=lambda e: e.energy_per_atom)[0]
            for comp_entries in entries_by_comp.values()
        ]
        pd = PhaseDiagram(reduced_entries)

        # Add back all entries, not just those on the hull
        pd_computed_data = pd.computed_data
        pd_computed_data["all_entries"] = entries
        new_pd = PhaseDiagram(
            entries, elements=pd.elements, computed_data=pd_computed_data
        )
        return new_pd


class PhaseDiagramDoc(BaseModel):
    """
    A phase diagram document
    """

    property_name: str = "phase_diagram"

    phase_diagram_id: str = Field(
        ...,
        description="Phase diagram ID consisting of the chemical system and thermo type",
    )

    chemsys: str = Field(
        ...,
        description="Dash-delimited string of elements in the material",
    )

    thermo_type: Union[ThermoType, RunType] = Field(
        ...,
        description="Functional types of calculations involved in the energy mixing scheme.",
    )

    phase_diagram: PhaseDiagram = Field(
        ...,
        description="Phase diagram for the chemical system.",
    )

    last_updated: datetime = Field(
        description="Timestamp for the most recent calculation update for this property",
        default_factory=datetime.utcnow,
    )



""" Core definition of a Task Document which represents a calculation from some program"""
from datetime import datetime
from typing import Union, Optional

from pydantic import Field

from emmet.core.base import EmmetBaseModel
from emmet.core.mpid import MPID, MPculeID


class BaseTaskDocument(EmmetBaseModel):
    """
    Definition of base Task Document
    """

    calc_code: str = Field(description="The calculation code used to compute this task")
    version: Optional[str] = Field(
        None, description="The version of the calculation code"
    )
    dir_name: Optional[str] = Field(None, description="The directory for this task")
    task_id: Optional[Union[MPID, MPculeID]] = Field(
        None, description="the Task ID For this document"
    )

    completed: bool = Field(False, description="Whether this calcuation completed")
    completed_at: Optional[datetime] = Field(
        None, description="Timestamp for when this task was completed"
    )
    last_updated: datetime = Field(
        default_factory=datetime.utcnow,
        description="Timestamp for when this task document was last updated",
    )



"""
Autogenerated Enums for VASP RunType, TaskType, and CalcType.

Do not edit this by hand. Edit generate.py or run_types.yaml instead.
"""
from emmet.core.utils import ValueEnum


class RunType(ValueEnum):
    """VASP calculation run types."""

    AM05 = "AM05"
    GGA = "GGA"
    PBE = "PBE"
    PBESol = "PBESol"
    RevPBE_PADE = "RevPBE+PADE"
    optB86b = "optB86b"
    optB88 = "optB88"
    optPBE = "optPBE"
    revPBE = "revPBE"
    B3LYP = "B3LYP"
    HF = "HF"
    HSE03 = "HSE03"
    HSE06 = "HSE06"
    HFCus = "HFCus"
    PBE0 = "PBE0"
    M06L = "M06L"
    MBJL = "MBJL"
    MS0 = "MS0"
    MS1 = "MS1"
    MS2 = "MS2"
    RTPSS = "RTPSS"
    SCAN = "SCAN"
    R2SCAN = "R2SCAN"
    TPSS = "TPSS"
    R2SCAN_rVV10 = "R2SCAN-rVV10"
    SCAN_rVV10 = "SCAN-rVV10"
    optB86b_vdW = "optB86b-vdW"
    optB88_vdW = "optB88-vdW"
    optPBE_vdW = "optPBE-vdW"
    rev_vdW_DF2 = "rev-vdW-DF2"
    revPBE_vdW = "revPBE-vdW"
    vdW_DF2 = "vdW-DF2"
    AM05_U = "AM05+U"
    GGA_U = "GGA+U"
    PBE_U = "PBE+U"
    PBESol_U = "PBESol+U"
    RevPBE_PADE_U = "RevPBE+PADE+U"
    optB86b_U = "optB86b+U"
    optB88_U = "optB88+U"
    optPBE_U = "optPBE+U"
    revPBE_U = "revPBE+U"
    B3LYP_U = "B3LYP+U"
    HF_U = "HF+U"
    HSE03_U = "HSE03+U"
    HSE06_U = "HSE06+U"
    HFCus_U = "HFCus+U"
    PBE0_U = "PBE0+U"
    M06L_U = "M06L+U"
    MBJL_U = "MBJL+U"
    MS0_U = "MS0+U"
    MS1_U = "MS1+U"
    MS2_U = "MS2+U"
    RTPSS_U = "RTPSS+U"
    SCAN_U = "SCAN+U"
    R2SCAN_U = "R2SCAN+U"
    TPSS_U = "TPSS+U"
    R2SCAN_rVV10_U = "R2SCAN-rVV10+U"
    SCAN_rVV10_U = "SCAN-rVV10+U"
    optB86b_vdW_U = "optB86b-vdW+U"
    optB88_vdW_U = "optB88-vdW+U"
    optPBE_vdW_U = "optPBE-vdW+U"
    rev_vdW_DF2_U = "rev-vdW-DF2+U"
    revPBE_vdW_U = "revPBE-vdW+U"
    vdW_DF2_U = "vdW-DF2+U"
    LDA = "LDA"
    LDA_U = "LDA+U"

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value.upper() == value.upper():
                return member


class TaskType(ValueEnum):
    """VASP calculation task types."""

    NSCF_Line = "NSCF Line"
    NSCF_Uniform = "NSCF Uniform"
    Dielectric = "Dielectric"
    DFPT = "DFPT"
    DFPT_Dielectric = "DFPT Dielectric"
    NMR_Nuclear_Shielding = "NMR Nuclear Shielding"
    NMR_Electric_Field_Gradient = "NMR Electric Field Gradient"
    Static = "Static"
    Structure_Optimization = "Structure Optimization"
    Deformation = "Deformation"
    Optic = "Optic"
    Molecular_Dynamics = "Molecular Dynamics"
    Unrecognized = "Unrecognized"


class CalcType(ValueEnum):
    """VASP calculation types."""

    AM05_NSCF_Line = "AM05 NSCF Line"
    AM05_NSCF_Uniform = "AM05 NSCF Uniform"
    AM05_Dielectric = "AM05 Dielectric"
    AM05_DFPT = "AM05 DFPT"
    AM05_DFPT_Dielectric = "AM05 DFPT Dielectric"
    AM05_NMR_Nuclear_Shielding = "AM05 NMR Nuclear Shielding"
    AM05_NMR_Electric_Field_Gradient = "AM05 NMR Electric Field Gradient"
    AM05_Static = "AM05 Static"
    AM05_Structure_Optimization = "AM05 Structure Optimization"
    AM05_Deformation = "AM05 Deformation"
    AM05_Optic = "AM05 Optic"
    AM05_Molecular_Dynamics = "AM05 Molecular Dynamics"
    AM05_Unrecognized = "AM05 Unrecognized"
    GGA_NSCF_Line = "GGA NSCF Line"
    GGA_NSCF_Uniform = "GGA NSCF Uniform"
    GGA_Dielectric = "GGA Dielectric"
    GGA_DFPT = "GGA DFPT"
    GGA_DFPT_Dielectric = "GGA DFPT Dielectric"
    GGA_NMR_Nuclear_Shielding = "GGA NMR Nuclear Shielding"
    GGA_NMR_Electric_Field_Gradient = "GGA NMR Electric Field Gradient"
    GGA_Static = "GGA Static"
    GGA_Structure_Optimization = "GGA Structure Optimization"
    GGA_Deformation = "GGA Deformation"
    GGA_Optic = "GGA Optic"
    GGA_Molecular_Dynamics = "GGA Molecular Dynamics"
    GGA_Unrecognized = "GGA Unrecognized"
    PBE_NSCF_Line = "PBE NSCF Line"
    PBE_NSCF_Uniform = "PBE NSCF Uniform"
    PBE_Dielectric = "PBE Dielectric"
    PBE_DFPT = "PBE DFPT"
    PBE_DFPT_Dielectric = "PBE DFPT Dielectric"
    PBE_NMR_Nuclear_Shielding = "PBE NMR Nuclear Shielding"
    PBE_NMR_Electric_Field_Gradient = "PBE NMR Electric Field Gradient"
    PBE_Static = "PBE Static"
    PBE_Structure_Optimization = "PBE Structure Optimization"
    PBE_Deformation = "PBE Deformation"
    PBE_Optic = "PBE Optic"
    PBE_Molecular_Dynamics = "PBE Molecular Dynamics"
    PBE_Unrecognized = "PBE Unrecognized"
    PBESol_NSCF_Line = "PBESol NSCF Line"
    PBESol_NSCF_Uniform = "PBESol NSCF Uniform"
    PBESol_Dielectric = "PBESol Dielectric"
    PBESol_DFPT = "PBESol DFPT"
    PBESol_DFPT_Dielectric = "PBESol DFPT Dielectric"
    PBESol_NMR_Nuclear_Shielding = "PBESol NMR Nuclear Shielding"
    PBESol_NMR_Electric_Field_Gradient = "PBESol NMR Electric Field Gradient"
    PBESol_Static = "PBESol Static"
    PBESol_Structure_Optimization = "PBESol Structure Optimization"
    PBESol_Deformation = "PBESol Deformation"
    PBESol_Optic = "PBESol Optic"
    PBESol_Molecular_Dynamics = "PBESol Molecular Dynamics"
    PBESol_Unrecognized = "PBESol Unrecognized"
    RevPBE_PADE_NSCF_Line = "RevPBE+PADE NSCF Line"
    RevPBE_PADE_NSCF_Uniform = "RevPBE+PADE NSCF Uniform"
    RevPBE_PADE_Dielectric = "RevPBE+PADE Dielectric"
    RevPBE_PADE_DFPT = "RevPBE+PADE DFPT"
    RevPBE_PADE_DFPT_Dielectric = "RevPBE+PADE DFPT Dielectric"
    RevPBE_PADE_NMR_Nuclear_Shielding = "RevPBE+PADE NMR Nuclear Shielding"
    RevPBE_PADE_NMR_Electric_Field_Gradient = "RevPBE+PADE NMR Electric Field Gradient"
    RevPBE_PADE_Static = "RevPBE+PADE Static"
    RevPBE_PADE_Structure_Optimization = "RevPBE+PADE Structure Optimization"
    RevPBE_PADE_Deformation = "RevPBE+PADE Deformation"
    RevPBE_PADE_Optic = "RevPBE+PADE Optic"
    RevPBE_PADE_Molecular_Dynamics = "RevPBE+PADE Molecular Dynamics"
    RevPBE_PADE_Unrecognized = "RevPBE+PADE Unrecognized"
    optB86b_NSCF_Line = "optB86b NSCF Line"
    optB86b_NSCF_Uniform = "optB86b NSCF Uniform"
    optB86b_Dielectric = "optB86b Dielectric"
    optB86b_DFPT = "optB86b DFPT"
    optB86b_DFPT_Dielectric = "optB86b DFPT Dielectric"
    optB86b_NMR_Nuclear_Shielding = "optB86b NMR Nuclear Shielding"
    optB86b_NMR_Electric_Field_Gradient = "optB86b NMR Electric Field Gradient"
    optB86b_Static = "optB86b Static"
    optB86b_Structure_Optimization = "optB86b Structure Optimization"
    optB86b_Deformation = "optB86b Deformation"
    optB86b_Optic = "optB86b Optic"
    optB86b_Molecular_Dynamics = "optB86b Molecular Dynamics"
    optB86b_Unrecognized = "optB86b Unrecognized"
    optB88_NSCF_Line = "optB88 NSCF Line"
    optB88_NSCF_Uniform = "optB88 NSCF Uniform"
    optB88_Dielectric = "optB88 Dielectric"
    optB88_DFPT = "optB88 DFPT"
    optB88_DFPT_Dielectric = "optB88 DFPT Dielectric"
    optB88_NMR_Nuclear_Shielding = "optB88 NMR Nuclear Shielding"
    optB88_NMR_Electric_Field_Gradient = "optB88 NMR Electric Field Gradient"
    optB88_Static = "optB88 Static"
    optB88_Structure_Optimization = "optB88 Structure Optimization"
    optB88_Deformation = "optB88 Deformation"
    optB88_Optic = "optB88 Optic"
    optB88_Molecular_Dynamics = "optB88 Molecular Dynamics"
    optB88_Unrecognized = "optB88 Unrecognized"
    optPBE_NSCF_Line = "optPBE NSCF Line"
    optPBE_NSCF_Uniform = "optPBE NSCF Uniform"
    optPBE_Dielectric = "optPBE Dielectric"
    optPBE_DFPT = "optPBE DFPT"
    optPBE_DFPT_Dielectric = "optPBE DFPT Dielectric"
    optPBE_NMR_Nuclear_Shielding = "optPBE NMR Nuclear Shielding"
    optPBE_NMR_Electric_Field_Gradient = "optPBE NMR Electric Field Gradient"
    optPBE_Static = "optPBE Static"
    optPBE_Structure_Optimization = "optPBE Structure Optimization"
    optPBE_Deformation = "optPBE Deformation"
    optPBE_Optic = "optPBE Optic"
    optPBE_Molecular_Dynamics = "optPBE Molecular Dynamics"
    optPBE_Unrecognized = "optPBE Unrecognized"
    revPBE_NSCF_Line = "revPBE NSCF Line"
    revPBE_NSCF_Uniform = "revPBE NSCF Uniform"
    revPBE_Dielectric = "revPBE Dielectric"
    revPBE_DFPT = "revPBE DFPT"
    revPBE_DFPT_Dielectric = "revPBE DFPT Dielectric"
    revPBE_NMR_Nuclear_Shielding = "revPBE NMR Nuclear Shielding"
    revPBE_NMR_Electric_Field_Gradient = "revPBE NMR Electric Field Gradient"
    revPBE_Static = "revPBE Static"
    revPBE_Structure_Optimization = "revPBE Structure Optimization"
    revPBE_Deformation = "revPBE Deformation"
    revPBE_Optic = "revPBE Optic"
    revPBE_Molecular_Dynamics = "revPBE Molecular Dynamics"
    revPBE_Unrecognized = "revPBE Unrecognized"
    B3LYP_NSCF_Line = "B3LYP NSCF Line"
    B3LYP_NSCF_Uniform = "B3LYP NSCF Uniform"
    B3LYP_Dielectric = "B3LYP Dielectric"
    B3LYP_DFPT = "B3LYP DFPT"
    B3LYP_DFPT_Dielectric = "B3LYP DFPT Dielectric"
    B3LYP_NMR_Nuclear_Shielding = "B3LYP NMR Nuclear Shielding"
    B3LYP_NMR_Electric_Field_Gradient = "B3LYP NMR Electric Field Gradient"
    B3LYP_Static = "B3LYP Static"
    B3LYP_Structure_Optimization = "B3LYP Structure Optimization"
    B3LYP_Deformation = "B3LYP Deformation"
    B3LYP_Optic = "B3LYP Optic"
    B3LYP_Molecular_Dynamics = "B3LYP Molecular Dynamics"
    B3LYP_Unrecognized = "B3LYP Unrecognized"
    HF_NSCF_Line = "HF NSCF Line"
    HF_NSCF_Uniform = "HF NSCF Uniform"
    HF_Dielectric = "HF Dielectric"
    HF_DFPT = "HF DFPT"
    HF_DFPT_Dielectric = "HF DFPT Dielectric"
    HF_NMR_Nuclear_Shielding = "HF NMR Nuclear Shielding"
    HF_NMR_Electric_Field_Gradient = "HF NMR Electric Field Gradient"
    HF_Static = "HF Static"
    HF_Structure_Optimization = "HF Structure Optimization"
    HF_Deformation = "HF Deformation"
    HF_Optic = "HF Optic"
    HF_Molecular_Dynamics = "HF Molecular Dynamics"
    HF_Unrecognized = "HF Unrecognized"
    HSE03_NSCF_Line = "HSE03 NSCF Line"
    HSE03_NSCF_Uniform = "HSE03 NSCF Uniform"
    HSE03_Dielectric = "HSE03 Dielectric"
    HSE03_DFPT = "HSE03 DFPT"
    HSE03_DFPT_Dielectric = "HSE03 DFPT Dielectric"
    HSE03_NMR_Nuclear_Shielding = "HSE03 NMR Nuclear Shielding"
    HSE03_NMR_Electric_Field_Gradient = "HSE03 NMR Electric Field Gradient"
    HSE03_Static = "HSE03 Static"
    HSE03_Structure_Optimization = "HSE03 Structure Optimization"
    HSE03_Deformation = "HSE03 Deformation"
    HSE03_Optic = "HSE03 Optic"
    HSE03_Molecular_Dynamics = "HSE03 Molecular Dynamics"
    HSE03_Unrecognized = "HSE03 Unrecognized"
    HSE06_NSCF_Line = "HSE06 NSCF Line"
    HSE06_NSCF_Uniform = "HSE06 NSCF Uniform"
    HSE06_Dielectric = "HSE06 Dielectric"
    HSE06_DFPT = "HSE06 DFPT"
    HSE06_DFPT_Dielectric = "HSE06 DFPT Dielectric"
    HSE06_NMR_Nuclear_Shielding = "HSE06 NMR Nuclear Shielding"
    HSE06_NMR_Electric_Field_Gradient = "HSE06 NMR Electric Field Gradient"
    HSE06_Static = "HSE06 Static"
    HSE06_Structure_Optimization = "HSE06 Structure Optimization"
    HSE06_Deformation = "HSE06 Deformation"
    HSE06_Optic = "HSE06 Optic"
    HSE06_Molecular_Dynamics = "HSE06 Molecular Dynamics"
    HSE06_Unrecognized = "HSE06 Unrecognized"
    PBE0_NSCF_Line = "PBE0 NSCF Line"
    PBE0_NSCF_Uniform = "PBE0 NSCF Uniform"
    PBE0_Dielectric = "PBE0 Dielectric"
    PBE0_DFPT = "PBE0 DFPT"
    PBE0_DFPT_Dielectric = "PBE0 DFPT Dielectric"
    PBE0_NMR_Nuclear_Shielding = "PBE0 NMR Nuclear Shielding"
    PBE0_NMR_Electric_Field_Gradient = "PBE0 NMR Electric Field Gradient"
    PBE0_Static = "PBE0 Static"
    PBE0_Structure_Optimization = "PBE0 Structure Optimization"
    PBE0_Deformation = "PBE0 Deformation"
    PBE0_Optic = "PBE0 Optic"
    PBE0_Molecular_Dynamics = "PBE0 Molecular Dynamics"
    PBE0_Unrecognized = "PBE0 Unrecognized"
    M06L_NSCF_Line = "M06L NSCF Line"
    M06L_NSCF_Uniform = "M06L NSCF Uniform"
    M06L_Dielectric = "M06L Dielectric"
    M06L_DFPT = "M06L DFPT"
    M06L_DFPT_Dielectric = "M06L DFPT Dielectric"
    M06L_NMR_Nuclear_Shielding = "M06L NMR Nuclear Shielding"
    M06L_NMR_Electric_Field_Gradient = "M06L NMR Electric Field Gradient"
    M06L_Static = "M06L Static"
    M06L_Structure_Optimization = "M06L Structure Optimization"
    M06L_Deformation = "M06L Deformation"
    M06L_Optic = "M06L Optic"
    M06L_Molecular_Dynamics = "M06L Molecular Dynamics"
    M06L_Unrecognized = "M06L Unrecognized"
    MBJL_NSCF_Line = "MBJL NSCF Line"
    MBJL_NSCF_Uniform = "MBJL NSCF Uniform"
    MBJL_Dielectric = "MBJL Dielectric"
    MBJL_DFPT = "MBJL DFPT"
    MBJL_DFPT_Dielectric = "MBJL DFPT Dielectric"
    MBJL_NMR_Nuclear_Shielding = "MBJL NMR Nuclear Shielding"
    MBJL_NMR_Electric_Field_Gradient = "MBJL NMR Electric Field Gradient"
    MBJL_Static = "MBJL Static"
    MBJL_Structure_Optimization = "MBJL Structure Optimization"
    MBJL_Deformation = "MBJL Deformation"
    MBJL_Optic = "MBJL Optic"
    MBJL_Molecular_Dynamics = "MBJL Molecular Dynamics"
    MBJL_Unrecognized = "MBJL Unrecognized"
    MS0_NSCF_Line = "MS0 NSCF Line"
    MS0_NSCF_Uniform = "MS0 NSCF Uniform"
    MS0_Dielectric = "MS0 Dielectric"
    MS0_DFPT = "MS0 DFPT"
    MS0_DFPT_Dielectric = "MS0 DFPT Dielectric"
    MS0_NMR_Nuclear_Shielding = "MS0 NMR Nuclear Shielding"
    MS0_NMR_Electric_Field_Gradient = "MS0 NMR Electric Field Gradient"
    MS0_Static = "MS0 Static"
    MS0_Structure_Optimization = "MS0 Structure Optimization"
    MS0_Deformation = "MS0 Deformation"
    MS0_Optic = "MS0 Optic"
    MS0_Molecular_Dynamics = "MS0 Molecular Dynamics"
    MS0_Unrecognized = "MS0 Unrecognized"
    MS1_NSCF_Line = "MS1 NSCF Line"
    MS1_NSCF_Uniform = "MS1 NSCF Uniform"
    MS1_Dielectric = "MS1 Dielectric"
    MS1_DFPT = "MS1 DFPT"
    MS1_DFPT_Dielectric = "MS1 DFPT Dielectric"
    MS1_NMR_Nuclear_Shielding = "MS1 NMR Nuclear Shielding"
    MS1_NMR_Electric_Field_Gradient = "MS1 NMR Electric Field Gradient"
    MS1_Static = "MS1 Static"
    MS1_Structure_Optimization = "MS1 Structure Optimization"
    MS1_Deformation = "MS1 Deformation"
    MS1_Optic = "MS1 Optic"
    MS1_Molecular_Dynamics = "MS1 Molecular Dynamics"
    MS1_Unrecognized = "MS1 Unrecognized"
    MS2_NSCF_Line = "MS2 NSCF Line"
    MS2_NSCF_Uniform = "MS2 NSCF Uniform"
    MS2_Dielectric = "MS2 Dielectric"
    MS2_DFPT = "MS2 DFPT"
    MS2_DFPT_Dielectric = "MS2 DFPT Dielectric"
    MS2_NMR_Nuclear_Shielding = "MS2 NMR Nuclear Shielding"
    MS2_NMR_Electric_Field_Gradient = "MS2 NMR Electric Field Gradient"
    MS2_Static = "MS2 Static"
    MS2_Structure_Optimization = "MS2 Structure Optimization"
    MS2_Deformation = "MS2 Deformation"
    MS2_Optic = "MS2 Optic"
    MS2_Molecular_Dynamics = "MS2 Molecular Dynamics"
    MS2_Unrecognized = "MS2 Unrecognized"
    RTPSS_NSCF_Line = "RTPSS NSCF Line"
    RTPSS_NSCF_Uniform = "RTPSS NSCF Uniform"
    RTPSS_Dielectric = "RTPSS Dielectric"
    RTPSS_DFPT = "RTPSS DFPT"
    RTPSS_DFPT_Dielectric = "RTPSS DFPT Dielectric"
    RTPSS_NMR_Nuclear_Shielding = "RTPSS NMR Nuclear Shielding"
    RTPSS_NMR_Electric_Field_Gradient = "RTPSS NMR Electric Field Gradient"
    RTPSS_Static = "RTPSS Static"
    RTPSS_Structure_Optimization = "RTPSS Structure Optimization"
    RTPSS_Deformation = "RTPSS Deformation"
    RTPSS_Optic = "RTPSS Optic"
    RTPSS_Molecular_Dynamics = "RTPSS Molecular Dynamics"
    RTPSS_Unrecognized = "RTPSS Unrecognized"
    SCAN_NSCF_Line = "SCAN NSCF Line"
    SCAN_NSCF_Uniform = "SCAN NSCF Uniform"
    SCAN_Dielectric = "SCAN Dielectric"
    SCAN_DFPT = "SCAN DFPT"
    SCAN_DFPT_Dielectric = "SCAN DFPT Dielectric"
    SCAN_NMR_Nuclear_Shielding = "SCAN NMR Nuclear Shielding"
    SCAN_NMR_Electric_Field_Gradient = "SCAN NMR Electric Field Gradient"
    SCAN_Static = "SCAN Static"
    SCAN_Structure_Optimization = "SCAN Structure Optimization"
    SCAN_Deformation = "SCAN Deformation"
    SCAN_Optic = "SCAN Optic"
    SCAN_Molecular_Dynamics = "SCAN Molecular Dynamics"
    SCAN_Unrecognized = "SCAN Unrecognized"
    R2SCAN_NSCF_Line = "R2SCAN NSCF Line"
    R2SCAN_NSCF_Uniform = "R2SCAN NSCF Uniform"
    R2SCAN_Dielectric = "R2SCAN Dielectric"
    R2SCAN_DFPT = "R2SCAN DFPT"
    R2SCAN_DFPT_Dielectric = "R2SCAN DFPT Dielectric"
    R2SCAN_NMR_Nuclear_Shielding = "R2SCAN NMR Nuclear Shielding"
    R2SCAN_NMR_Electric_Field_Gradient = "R2SCAN NMR Electric Field Gradient"
    R2SCAN_Static = "R2SCAN Static"
    R2SCAN_Structure_Optimization = "R2SCAN Structure Optimization"
    R2SCAN_Deformation = "R2SCAN Deformation"
    R2SCAN_Optic = "R2SCAN Optic"
    R2SCAN_Molecular_Dynamics = "R2SCAN Molecular Dynamics"
    R2SCAN_Unrecognized = "R2SCAN Unrecognized"
    TPSS_NSCF_Line = "TPSS NSCF Line"
    TPSS_NSCF_Uniform = "TPSS NSCF Uniform"
    TPSS_Dielectric = "TPSS Dielectric"
    TPSS_DFPT = "TPSS DFPT"
    TPSS_DFPT_Dielectric = "TPSS DFPT Dielectric"
    TPSS_NMR_Nuclear_Shielding = "TPSS NMR Nuclear Shielding"
    TPSS_NMR_Electric_Field_Gradient = "TPSS NMR Electric Field Gradient"
    TPSS_Static = "TPSS Static"
    TPSS_Structure_Optimization = "TPSS Structure Optimization"
    TPSS_Deformation = "TPSS Deformation"
    TPSS_Optic = "TPSS Optic"
    TPSS_Molecular_Dynamics = "TPSS Molecular Dynamics"
    TPSS_Unrecognized = "TPSS Unrecognized"
    R2SCAN_rVV10_NSCF_Line = "R2SCAN-rVV10 NSCF Line"
    R2SCAN_rVV10_NSCF_Uniform = "R2SCAN-rVV10 NSCF Uniform"
    R2SCAN_rVV10_Dielectric = "R2SCAN-rVV10 Dielectric"
    R2SCAN_rVV10_DFPT = "R2SCAN-rVV10 DFPT"
    R2SCAN_rVV10_DFPT_Dielectric = "R2SCAN-rVV10 DFPT Dielectric"
    R2SCAN_rVV10_NMR_Nuclear_Shielding = "R2SCAN-rVV10 NMR Nuclear Shielding"
    R2SCAN_rVV10_NMR_Electric_Field_Gradient = (
        "R2SCAN-rVV10 NMR Electric Field Gradient"
    )
    R2SCAN_rVV10_Static = "R2SCAN-rVV10 Static"
    R2SCAN_rVV10_Structure_Optimization = "R2SCAN-rVV10 Structure Optimization"
    R2SCAN_rVV10_Deformation = "R2SCAN-rVV10 Deformation"
    R2SCAN_rVV10_Optic = "R2SCAN-rVV10 Optic"
    R2SCAN_rVV10_Molecular_Dynamics = "R2SCAN-rVV10 Molecular Dynamics"
    R2SCAN_rVV10_Unrecognized = "R2SCAN-rVV10 Unrecognized"
    SCAN_rVV10_NSCF_Line = "SCAN-rVV10 NSCF Line"
    SCAN_rVV10_NSCF_Uniform = "SCAN-rVV10 NSCF Uniform"
    SCAN_rVV10_Dielectric = "SCAN-rVV10 Dielectric"
    SCAN_rVV10_DFPT = "SCAN-rVV10 DFPT"
    SCAN_rVV10_DFPT_Dielectric = "SCAN-rVV10 DFPT Dielectric"
    SCAN_rVV10_NMR_Nuclear_Shielding = "SCAN-rVV10 NMR Nuclear Shielding"
    SCAN_rVV10_NMR_Electric_Field_Gradient = "SCAN-rVV10 NMR Electric Field Gradient"
    SCAN_rVV10_Static = "SCAN-rVV10 Static"
    SCAN_rVV10_Structure_Optimization = "SCAN-rVV10 Structure Optimization"
    SCAN_rVV10_Deformation = "SCAN-rVV10 Deformation"
    SCAN_rVV10_Optic = "SCAN-rVV10 Optic"
    SCAN_rVV10_Molecular_Dynamics = "SCAN-rVV10 Molecular Dynamics"
    SCAN_rVV10_Unrecognized = "SCAN-rVV10 Unrecognized"
    optB86b_vdW_NSCF_Line = "optB86b-vdW NSCF Line"
    optB86b_vdW_NSCF_Uniform = "optB86b-vdW NSCF Uniform"
    optB86b_vdW_Dielectric = "optB86b-vdW Dielectric"
    optB86b_vdW_DFPT = "optB86b-vdW DFPT"
    optB86b_vdW_DFPT_Dielectric = "optB86b-vdW DFPT Dielectric"
    optB86b_vdW_NMR_Nuclear_Shielding = "optB86b-vdW NMR Nuclear Shielding"
    optB86b_vdW_NMR_Electric_Field_Gradient = "optB86b-vdW NMR Electric Field Gradient"
    optB86b_vdW_Static = "optB86b-vdW Static"
    optB86b_vdW_Structure_Optimization = "optB86b-vdW Structure Optimization"
    optB86b_vdW_Deformation = "optB86b-vdW Deformation"
    optB86b_vdW_Optic = "optB86b-vdW Optic"
    optB86b_vdW_Molecular_Dynamics = "optB86b-vdW Molecular Dynamics"
    optB86b_vdW_Unrecognized = "optB86b-vdW Unrecognized"
    optB88_vdW_NSCF_Line = "optB88-vdW NSCF Line"
    optB88_vdW_NSCF_Uniform = "optB88-vdW NSCF Uniform"
    optB88_vdW_Dielectric = "optB88-vdW Dielectric"
    optB88_vdW_DFPT = "optB88-vdW DFPT"
    optB88_vdW_DFPT_Dielectric = "optB88-vdW DFPT Dielectric"
    optB88_vdW_NMR_Nuclear_Shielding = "optB88-vdW NMR Nuclear Shielding"
    optB88_vdW_NMR_Electric_Field_Gradient = "optB88-vdW NMR Electric Field Gradient"
    optB88_vdW_Static = "optB88-vdW Static"
    optB88_vdW_Structure_Optimization = "optB88-vdW Structure Optimization"
    optB88_vdW_Deformation = "optB88-vdW Deformation"
    optB88_vdW_Optic = "optB88-vdW Optic"
    optB88_vdW_Molecular_Dynamics = "optB88-vdW Molecular Dynamics"
    optB88_vdW_Unrecognized = "optB88-vdW Unrecognized"
    optPBE_vdW_NSCF_Line = "optPBE-vdW NSCF Line"
    optPBE_vdW_NSCF_Uniform = "optPBE-vdW NSCF Uniform"
    optPBE_vdW_Dielectric = "optPBE-vdW Dielectric"
    optPBE_vdW_DFPT = "optPBE-vdW DFPT"
    optPBE_vdW_DFPT_Dielectric = "optPBE-vdW DFPT Dielectric"
    optPBE_vdW_NMR_Nuclear_Shielding = "optPBE-vdW NMR Nuclear Shielding"
    optPBE_vdW_NMR_Electric_Field_Gradient = "optPBE-vdW NMR Electric Field Gradient"
    optPBE_vdW_Static = "optPBE-vdW Static"
    optPBE_vdW_Structure_Optimization = "optPBE-vdW Structure Optimization"
    optPBE_vdW_Deformation = "optPBE-vdW Deformation"
    optPBE_vdW_Optic = "optPBE-vdW Optic"
    optPBE_vdW_Molecular_Dynamics = "optPBE-vdW Molecular Dynamics"
    optPBE_vdW_Unrecognized = "optPBE-vdW Unrecognized"
    rev_vdW_DF2_NSCF_Line = "rev-vdW-DF2 NSCF Line"
    rev_vdW_DF2_NSCF_Uniform = "rev-vdW-DF2 NSCF Uniform"
    rev_vdW_DF2_Dielectric = "rev-vdW-DF2 Dielectric"
    rev_vdW_DF2_DFPT = "rev-vdW-DF2 DFPT"
    rev_vdW_DF2_DFPT_Dielectric = "rev-vdW-DF2 DFPT Dielectric"
    rev_vdW_DF2_NMR_Nuclear_Shielding = "rev-vdW-DF2 NMR Nuclear Shielding"
    rev_vdW_DF2_NMR_Electric_Field_Gradient = "rev-vdW-DF2 NMR Electric Field Gradient"
    rev_vdW_DF2_Static = "rev-vdW-DF2 Static"
    rev_vdW_DF2_Structure_Optimization = "rev-vdW-DF2 Structure Optimization"
    rev_vdW_DF2_Deformation = "rev-vdW-DF2 Deformation"
    rev_vdW_DF2_Optic = "rev-vdW-DF2 Optic"
    rev_vdW_DF2_Molecular_Dynamics = "rev-vdW-DF2 Molecular Dynamics"
    rev_vdW_DF2_Unrecognized = "rev-vdW-DF2 Unrecognized"
    revPBE_vdW_NSCF_Line = "revPBE-vdW NSCF Line"
    revPBE_vdW_NSCF_Uniform = "revPBE-vdW NSCF Uniform"
    revPBE_vdW_Dielectric = "revPBE-vdW Dielectric"
    revPBE_vdW_DFPT = "revPBE-vdW DFPT"
    revPBE_vdW_DFPT_Dielectric = "revPBE-vdW DFPT Dielectric"
    revPBE_vdW_NMR_Nuclear_Shielding = "revPBE-vdW NMR Nuclear Shielding"
    revPBE_vdW_NMR_Electric_Field_Gradient = "revPBE-vdW NMR Electric Field Gradient"
    revPBE_vdW_Static = "revPBE-vdW Static"
    revPBE_vdW_Structure_Optimization = "revPBE-vdW Structure Optimization"
    revPBE_vdW_Deformation = "revPBE-vdW Deformation"
    revPBE_vdW_Optic = "revPBE-vdW Optic"
    revPBE_vdW_Molecular_Dynamics = "revPBE-vdW Molecular Dynamics"
    revPBE_vdW_Unrecognized = "revPBE-vdW Unrecognized"
    vdW_DF2_NSCF_Line = "vdW-DF2 NSCF Line"
    vdW_DF2_NSCF_Uniform = "vdW-DF2 NSCF Uniform"
    vdW_DF2_Dielectric = "vdW-DF2 Dielectric"
    vdW_DF2_DFPT = "vdW-DF2 DFPT"
    vdW_DF2_DFPT_Dielectric = "vdW-DF2 DFPT Dielectric"
    vdW_DF2_NMR_Nuclear_Shielding = "vdW-DF2 NMR Nuclear Shielding"
    vdW_DF2_NMR_Electric_Field_Gradient = "vdW-DF2 NMR Electric Field Gradient"
    vdW_DF2_Static = "vdW-DF2 Static"
    vdW_DF2_Structure_Optimization = "vdW-DF2 Structure Optimization"
    vdW_DF2_Deformation = "vdW-DF2 Deformation"
    vdW_DF2_Optic = "vdW-DF2 Optic"
    vdW_DF2_Molecular_Dynamics = "vdW-DF2 Molecular Dynamics"
    vdW_DF2_Unrecognized = "vdW-DF2 Unrecognized"
    AM05_U_NSCF_Line = "AM05+U NSCF Line"
    AM05_U_NSCF_Uniform = "AM05+U NSCF Uniform"
    AM05_U_Dielectric = "AM05+U Dielectric"
    AM05_U_DFPT = "AM05+U DFPT"
    AM05_U_DFPT_Dielectric = "AM05+U DFPT Dielectric"
    AM05_U_NMR_Nuclear_Shielding = "AM05+U NMR Nuclear Shielding"
    AM05_U_NMR_Electric_Field_Gradient = "AM05+U NMR Electric Field Gradient"
    AM05_U_Static = "AM05+U Static"
    AM05_U_Structure_Optimization = "AM05+U Structure Optimization"
    AM05_U_Deformation = "AM05+U Deformation"
    AM05_U_Optic = "AM05+U Optic"
    AM05_U_Molecular_Dynamics = "AM05+U Molecular Dynamics"
    AM05_U_Unrecognized = "AM05+U Unrecognized"
    GGA_U_NSCF_Line = "GGA+U NSCF Line"
    GGA_U_NSCF_Uniform = "GGA+U NSCF Uniform"
    GGA_U_Dielectric = "GGA+U Dielectric"
    GGA_U_DFPT = "GGA+U DFPT"
    GGA_U_DFPT_Dielectric = "GGA+U DFPT Dielectric"
    GGA_U_NMR_Nuclear_Shielding = "GGA+U NMR Nuclear Shielding"
    GGA_U_NMR_Electric_Field_Gradient = "GGA+U NMR Electric Field Gradient"
    GGA_U_Static = "GGA+U Static"
    GGA_U_Structure_Optimization = "GGA+U Structure Optimization"
    GGA_U_Deformation = "GGA+U Deformation"
    GGA_U_Optic = "GGA+U Optic"
    GGA_U_Molecular_Dynamics = "GGA+U Molecular Dynamics"
    GGA_U_Unrecognized = "GGA+U Unrecognized"
    PBE_U_NSCF_Line = "PBE+U NSCF Line"
    PBE_U_NSCF_Uniform = "PBE+U NSCF Uniform"
    PBE_U_Dielectric = "PBE+U Dielectric"
    PBE_U_DFPT = "PBE+U DFPT"
    PBE_U_DFPT_Dielectric = "PBE+U DFPT Dielectric"
    PBE_U_NMR_Nuclear_Shielding = "PBE+U NMR Nuclear Shielding"
    PBE_U_NMR_Electric_Field_Gradient = "PBE+U NMR Electric Field Gradient"
    PBE_U_Static = "PBE+U Static"
    PBE_U_Structure_Optimization = "PBE+U Structure Optimization"
    PBE_U_Deformation = "PBE+U Deformation"
    PBE_U_Optic = "PBE+U Optic"
    PBE_U_Molecular_Dynamics = "PBE+U Molecular Dynamics"
    PBE_U_Unrecognized = "PBE+U Unrecognized"
    PBESol_U_NSCF_Line = "PBESol+U NSCF Line"
    PBESol_U_NSCF_Uniform = "PBESol+U NSCF Uniform"
    PBESol_U_Dielectric = "PBESol+U Dielectric"
    PBESol_U_DFPT = "PBESol+U DFPT"
    PBESol_U_DFPT_Dielectric = "PBESol+U DFPT Dielectric"
    PBESol_U_NMR_Nuclear_Shielding = "PBESol+U NMR Nuclear Shielding"
    PBESol_U_NMR_Electric_Field_Gradient = "PBESol+U NMR Electric Field Gradient"
    PBESol_U_Static = "PBESol+U Static"
    PBESol_U_Structure_Optimization = "PBESol+U Structure Optimization"
    PBESol_U_Deformation = "PBESol+U Deformation"
    PBESol_U_Optic = "PBESol+U Optic"
    PBESol_U_Molecular_Dynamics = "PBESol+U Molecular Dynamics"
    PBESol_U_Unrecognized = "PBESol+U Unrecognized"
    RevPBE_PADE_U_NSCF_Line = "RevPBE+PADE+U NSCF Line"
    RevPBE_PADE_U_NSCF_Uniform = "RevPBE+PADE+U NSCF Uniform"
    RevPBE_PADE_U_Dielectric = "RevPBE+PADE+U Dielectric"
    RevPBE_PADE_U_DFPT = "RevPBE+PADE+U DFPT"
    RevPBE_PADE_U_DFPT_Dielectric = "RevPBE+PADE+U DFPT Dielectric"
    RevPBE_PADE_U_NMR_Nuclear_Shielding = "RevPBE+PADE+U NMR Nuclear Shielding"
    RevPBE_PADE_U_NMR_Electric_Field_Gradient = (
        "RevPBE+PADE+U NMR Electric Field Gradient"
    )
    RevPBE_PADE_U_Static = "RevPBE+PADE+U Static"
    RevPBE_PADE_U_Structure_Optimization = "RevPBE+PADE+U Structure Optimization"
    RevPBE_PADE_U_Deformation = "RevPBE+PADE+U Deformation"
    RevPBE_PADE_U_Optic = "RevPBE+PADE+U Optic"
    RevPBE_PADE_U_Molecular_Dynamics = "RevPBE+PADE+U Molecular Dynamics"
    RevPBE_PADE_U_Unrecognized = "RevPBE+PADE+U Unrecognized"
    optB86b_U_NSCF_Line = "optB86b+U NSCF Line"
    optB86b_U_NSCF_Uniform = "optB86b+U NSCF Uniform"
    optB86b_U_Dielectric = "optB86b+U Dielectric"
    optB86b_U_DFPT = "optB86b+U DFPT"
    optB86b_U_DFPT_Dielectric = "optB86b+U DFPT Dielectric"
    optB86b_U_NMR_Nuclear_Shielding = "optB86b+U NMR Nuclear Shielding"
    optB86b_U_NMR_Electric_Field_Gradient = "optB86b+U NMR Electric Field Gradient"
    optB86b_U_Static = "optB86b+U Static"
    optB86b_U_Structure_Optimization = "optB86b+U Structure Optimization"
    optB86b_U_Deformation = "optB86b+U Deformation"
    optB86b_U_Optic = "optB86b+U Optic"
    optB86b_U_Molecular_Dynamics = "optB86b+U Molecular Dynamics"
    optB86b_U_Unrecognized = "optB86b+U Unrecognized"
    optB88_U_NSCF_Line = "optB88+U NSCF Line"
    optB88_U_NSCF_Uniform = "optB88+U NSCF Uniform"
    optB88_U_Dielectric = "optB88+U Dielectric"
    optB88_U_DFPT = "optB88+U DFPT"
    optB88_U_DFPT_Dielectric = "optB88+U DFPT Dielectric"
    optB88_U_NMR_Nuclear_Shielding = "optB88+U NMR Nuclear Shielding"
    optB88_U_NMR_Electric_Field_Gradient = "optB88+U NMR Electric Field Gradient"
    optB88_U_Static = "optB88+U Static"
    optB88_U_Structure_Optimization = "optB88+U Structure Optimization"
    optB88_U_Deformation = "optB88+U Deformation"
    optB88_U_Optic = "optB88+U Optic"
    optB88_U_Molecular_Dynamics = "optB88+U Molecular Dynamics"
    optB88_U_Unrecognized = "optB88+U Unrecognized"
    optPBE_U_NSCF_Line = "optPBE+U NSCF Line"
    optPBE_U_NSCF_Uniform = "optPBE+U NSCF Uniform"
    optPBE_U_Dielectric = "optPBE+U Dielectric"
    optPBE_U_DFPT = "optPBE+U DFPT"
    optPBE_U_DFPT_Dielectric = "optPBE+U DFPT Dielectric"
    optPBE_U_NMR_Nuclear_Shielding = "optPBE+U NMR Nuclear Shielding"
    optPBE_U_NMR_Electric_Field_Gradient = "optPBE+U NMR Electric Field Gradient"
    optPBE_U_Static = "optPBE+U Static"
    optPBE_U_Structure_Optimization = "optPBE+U Structure Optimization"
    optPBE_U_Deformation = "optPBE+U Deformation"
    optPBE_U_Optic = "optPBE+U Optic"
    optPBE_U_Molecular_Dynamics = "optPBE+U Molecular Dynamics"
    optPBE_U_Unrecognized = "optPBE+U Unrecognized"
    revPBE_U_NSCF_Line = "revPBE+U NSCF Line"
    revPBE_U_NSCF_Uniform = "revPBE+U NSCF Uniform"
    revPBE_U_Dielectric = "revPBE+U Dielectric"
    revPBE_U_DFPT = "revPBE+U DFPT"
    revPBE_U_DFPT_Dielectric = "revPBE+U DFPT Dielectric"
    revPBE_U_NMR_Nuclear_Shielding = "revPBE+U NMR Nuclear Shielding"
    revPBE_U_NMR_Electric_Field_Gradient = "revPBE+U NMR Electric Field Gradient"
    revPBE_U_Static = "revPBE+U Static"
    revPBE_U_Structure_Optimization = "revPBE+U Structure Optimization"
    revPBE_U_Deformation = "revPBE+U Deformation"
    revPBE_U_Optic = "revPBE+U Optic"
    revPBE_U_Molecular_Dynamics = "revPBE+U Molecular Dynamics"
    revPBE_U_Unrecognized = "revPBE+U Unrecognized"
    B3LYP_U_NSCF_Line = "B3LYP+U NSCF Line"
    B3LYP_U_NSCF_Uniform = "B3LYP+U NSCF Uniform"
    B3LYP_U_Dielectric = "B3LYP+U Dielectric"
    B3LYP_U_DFPT = "B3LYP+U DFPT"
    B3LYP_U_DFPT_Dielectric = "B3LYP+U DFPT Dielectric"
    B3LYP_U_NMR_Nuclear_Shielding = "B3LYP+U NMR Nuclear Shielding"
    B3LYP_U_NMR_Electric_Field_Gradient = "B3LYP+U NMR Electric Field Gradient"
    B3LYP_U_Static = "B3LYP+U Static"
    B3LYP_U_Structure_Optimization = "B3LYP+U Structure Optimization"
    B3LYP_U_Deformation = "B3LYP+U Deformation"
    B3LYP_U_Optic = "B3LYP+U Optic"
    B3LYP_U_Molecular_Dynamics = "B3LYP+U Molecular Dynamics"
    B3LYP_U_Unrecognized = "B3LYP+U Unrecognized"
    HF_U_NSCF_Line = "HF+U NSCF Line"
    HF_U_NSCF_Uniform = "HF+U NSCF Uniform"
    HF_U_Dielectric = "HF+U Dielectric"
    HF_U_DFPT = "HF+U DFPT"
    HF_U_DFPT_Dielectric = "HF+U DFPT Dielectric"
    HF_U_NMR_Nuclear_Shielding = "HF+U NMR Nuclear Shielding"
    HF_U_NMR_Electric_Field_Gradient = "HF+U NMR Electric Field Gradient"
    HF_U_Static = "HF+U Static"
    HF_U_Structure_Optimization = "HF+U Structure Optimization"
    HF_U_Deformation = "HF+U Deformation"
    HF_U_Optic = "HF+U Optic"
    HF_U_Molecular_Dynamics = "HF+U Molecular Dynamics"
    HF_U_Unrecognized = "HF+U Unrecognized"
    HSE03_U_NSCF_Line = "HSE03+U NSCF Line"
    HSE03_U_NSCF_Uniform = "HSE03+U NSCF Uniform"
    HSE03_U_Dielectric = "HSE03+U Dielectric"
    HSE03_U_DFPT = "HSE03+U DFPT"
    HSE03_U_DFPT_Dielectric = "HSE03+U DFPT Dielectric"
    HSE03_U_NMR_Nuclear_Shielding = "HSE03+U NMR Nuclear Shielding"
    HSE03_U_NMR_Electric_Field_Gradient = "HSE03+U NMR Electric Field Gradient"
    HSE03_U_Static = "HSE03+U Static"
    HSE03_U_Structure_Optimization = "HSE03+U Structure Optimization"
    HSE03_U_Deformation = "HSE03+U Deformation"
    HSE03_U_Optic = "HSE03+U Optic"
    HSE03_U_Molecular_Dynamics = "HSE03+U Molecular Dynamics"
    HSE03_U_Unrecognized = "HSE03+U Unrecognized"
    HSE06_U_NSCF_Line = "HSE06+U NSCF Line"
    HSE06_U_NSCF_Uniform = "HSE06+U NSCF Uniform"
    HSE06_U_Dielectric = "HSE06+U Dielectric"
    HSE06_U_DFPT = "HSE06+U DFPT"
    HSE06_U_DFPT_Dielectric = "HSE06+U DFPT Dielectric"
    HSE06_U_NMR_Nuclear_Shielding = "HSE06+U NMR Nuclear Shielding"
    HSE06_U_NMR_Electric_Field_Gradient = "HSE06+U NMR Electric Field Gradient"
    HSE06_U_Static = "HSE06+U Static"
    HSE06_U_Structure_Optimization = "HSE06+U Structure Optimization"
    HSE06_U_Deformation = "HSE06+U Deformation"
    HSE06_U_Optic = "HSE06+U Optic"
    HSE06_U_Molecular_Dynamics = "HSE06+U Molecular Dynamics"
    HSE06_U_Unrecognized = "HSE06+U Unrecognized"
    PBE0_U_NSCF_Line = "PBE0+U NSCF Line"
    PBE0_U_NSCF_Uniform = "PBE0+U NSCF Uniform"
    PBE0_U_Dielectric = "PBE0+U Dielectric"
    PBE0_U_DFPT = "PBE0+U DFPT"
    PBE0_U_DFPT_Dielectric = "PBE0+U DFPT Dielectric"
    PBE0_U_NMR_Nuclear_Shielding = "PBE0+U NMR Nuclear Shielding"
    PBE0_U_NMR_Electric_Field_Gradient = "PBE0+U NMR Electric Field Gradient"
    PBE0_U_Static = "PBE0+U Static"
    PBE0_U_Structure_Optimization = "PBE0+U Structure Optimization"
    PBE0_U_Deformation = "PBE0+U Deformation"
    PBE0_U_Optic = "PBE0+U Optic"
    PBE0_U_Molecular_Dynamics = "PBE0+U Molecular Dynamics"
    PBE0_U_Unrecognized = "PBE0+U Unrecognized"
    M06L_U_NSCF_Line = "M06L+U NSCF Line"
    M06L_U_NSCF_Uniform = "M06L+U NSCF Uniform"
    M06L_U_Dielectric = "M06L+U Dielectric"
    M06L_U_DFPT = "M06L+U DFPT"
    M06L_U_DFPT_Dielectric = "M06L+U DFPT Dielectric"
    M06L_U_NMR_Nuclear_Shielding = "M06L+U NMR Nuclear Shielding"
    M06L_U_NMR_Electric_Field_Gradient = "M06L+U NMR Electric Field Gradient"
    M06L_U_Static = "M06L+U Static"
    M06L_U_Structure_Optimization = "M06L+U Structure Optimization"
    M06L_U_Deformation = "M06L+U Deformation"
    M06L_U_Optic = "M06L+U Optic"
    M06L_U_Molecular_Dynamics = "M06L+U Molecular Dynamics"
    M06L_U_Unrecognized = "M06L+U Unrecognized"
    MBJL_U_NSCF_Line = "MBJL+U NSCF Line"
    MBJL_U_NSCF_Uniform = "MBJL+U NSCF Uniform"
    MBJL_U_Dielectric = "MBJL+U Dielectric"
    MBJL_U_DFPT = "MBJL+U DFPT"
    MBJL_U_DFPT_Dielectric = "MBJL+U DFPT Dielectric"
    MBJL_U_NMR_Nuclear_Shielding = "MBJL+U NMR Nuclear Shielding"
    MBJL_U_NMR_Electric_Field_Gradient = "MBJL+U NMR Electric Field Gradient"
    MBJL_U_Static = "MBJL+U Static"
    MBJL_U_Structure_Optimization = "MBJL+U Structure Optimization"
    MBJL_U_Deformation = "MBJL+U Deformation"
    MBJL_U_Optic = "MBJL+U Optic"
    MBJL_U_Molecular_Dynamics = "MBJL+U Molecular Dynamics"
    MBJL_U_Unrecognized = "MBJL+U Unrecognized"
    MS0_U_NSCF_Line = "MS0+U NSCF Line"
    MS0_U_NSCF_Uniform = "MS0+U NSCF Uniform"
    MS0_U_Dielectric = "MS0+U Dielectric"
    MS0_U_DFPT = "MS0+U DFPT"
    MS0_U_DFPT_Dielectric = "MS0+U DFPT Dielectric"
    MS0_U_NMR_Nuclear_Shielding = "MS0+U NMR Nuclear Shielding"
    MS0_U_NMR_Electric_Field_Gradient = "MS0+U NMR Electric Field Gradient"
    MS0_U_Static = "MS0+U Static"
    MS0_U_Structure_Optimization = "MS0+U Structure Optimization"
    MS0_U_Deformation = "MS0+U Deformation"
    MS0_U_Optic = "MS0+U Optic"
    MS0_U_Molecular_Dynamics = "MS0+U Molecular Dynamics"
    MS0_U_Unrecognized = "MS0+U Unrecognized"
    MS1_U_NSCF_Line = "MS1+U NSCF Line"
    MS1_U_NSCF_Uniform = "MS1+U NSCF Uniform"
    MS1_U_Dielectric = "MS1+U Dielectric"
    MS1_U_DFPT = "MS1+U DFPT"
    MS1_U_DFPT_Dielectric = "MS1+U DFPT Dielectric"
    MS1_U_NMR_Nuclear_Shielding = "MS1+U NMR Nuclear Shielding"
    MS1_U_NMR_Electric_Field_Gradient = "MS1+U NMR Electric Field Gradient"
    MS1_U_Static = "MS1+U Static"
    MS1_U_Structure_Optimization = "MS1+U Structure Optimization"
    MS1_U_Deformation = "MS1+U Deformation"
    MS1_U_Optic = "MS1+U Optic"
    MS1_U_Molecular_Dynamics = "MS1+U Molecular Dynamics"
    MS1_U_Unrecognized = "MS1+U Unrecognized"
    MS2_U_NSCF_Line = "MS2+U NSCF Line"
    MS2_U_NSCF_Uniform = "MS2+U NSCF Uniform"
    MS2_U_Dielectric = "MS2+U Dielectric"
    MS2_U_DFPT = "MS2+U DFPT"
    MS2_U_DFPT_Dielectric = "MS2+U DFPT Dielectric"
    MS2_U_NMR_Nuclear_Shielding = "MS2+U NMR Nuclear Shielding"
    MS2_U_NMR_Electric_Field_Gradient = "MS2+U NMR Electric Field Gradient"
    MS2_U_Static = "MS2+U Static"
    MS2_U_Structure_Optimization = "MS2+U Structure Optimization"
    MS2_U_Deformation = "MS2+U Deformation"
    MS2_U_Optic = "MS2+U Optic"
    MS2_U_Molecular_Dynamics = "MS2+U Molecular Dynamics"
    MS2_U_Unrecognized = "MS2+U Unrecognized"
    RTPSS_U_NSCF_Line = "RTPSS+U NSCF Line"
    RTPSS_U_NSCF_Uniform = "RTPSS+U NSCF Uniform"
    RTPSS_U_Dielectric = "RTPSS+U Dielectric"
    RTPSS_U_DFPT = "RTPSS+U DFPT"
    RTPSS_U_DFPT_Dielectric = "RTPSS+U DFPT Dielectric"
    RTPSS_U_NMR_Nuclear_Shielding = "RTPSS+U NMR Nuclear Shielding"
    RTPSS_U_NMR_Electric_Field_Gradient = "RTPSS+U NMR Electric Field Gradient"
    RTPSS_U_Static = "RTPSS+U Static"
    RTPSS_U_Structure_Optimization = "RTPSS+U Structure Optimization"
    RTPSS_U_Deformation = "RTPSS+U Deformation"
    RTPSS_U_Optic = "RTPSS+U Optic"
    RTPSS_U_Molecular_Dynamics = "RTPSS+U Molecular Dynamics"
    RTPSS_U_Unrecognized = "RTPSS+U Unrecognized"
    SCAN_U_NSCF_Line = "SCAN+U NSCF Line"
    SCAN_U_NSCF_Uniform = "SCAN+U NSCF Uniform"
    SCAN_U_Dielectric = "SCAN+U Dielectric"
    SCAN_U_DFPT = "SCAN+U DFPT"
    SCAN_U_DFPT_Dielectric = "SCAN+U DFPT Dielectric"
    SCAN_U_NMR_Nuclear_Shielding = "SCAN+U NMR Nuclear Shielding"
    SCAN_U_NMR_Electric_Field_Gradient = "SCAN+U NMR Electric Field Gradient"
    SCAN_U_Static = "SCAN+U Static"
    SCAN_U_Structure_Optimization = "SCAN+U Structure Optimization"
    SCAN_U_Deformation = "SCAN+U Deformation"
    SCAN_U_Optic = "SCAN+U Optic"
    SCAN_U_Molecular_Dynamics = "SCAN+U Molecular Dynamics"
    SCAN_U_Unrecognized = "SCAN+U Unrecognized"
    R2SCAN_U_NSCF_Line = "R2SCAN+U NSCF Line"
    R2SCAN_U_NSCF_Uniform = "R2SCAN+U NSCF Uniform"
    R2SCAN_U_Dielectric = "R2SCAN+U Dielectric"
    R2SCAN_U_DFPT = "R2SCAN+U DFPT"
    R2SCAN_U_DFPT_Dielectric = "R2SCAN+U DFPT Dielectric"
    R2SCAN_U_NMR_Nuclear_Shielding = "R2SCAN+U NMR Nuclear Shielding"
    R2SCAN_U_NMR_Electric_Field_Gradient = "R2SCAN+U NMR Electric Field Gradient"
    R2SCAN_U_Static = "R2SCAN+U Static"
    R2SCAN_U_Structure_Optimization = "R2SCAN+U Structure Optimization"
    R2SCAN_U_Deformation = "R2SCAN+U Deformation"
    R2SCAN_U_Optic = "R2SCAN+U Optic"
    R2SCAN_U_Molecular_Dynamics = "R2SCAN+U Molecular Dynamics"
    R2SCAN_U_Unrecognized = "R2SCAN+U Unrecognized"
    TPSS_U_NSCF_Line = "TPSS+U NSCF Line"
    TPSS_U_NSCF_Uniform = "TPSS+U NSCF Uniform"
    TPSS_U_Dielectric = "TPSS+U Dielectric"
    TPSS_U_DFPT = "TPSS+U DFPT"
    TPSS_U_DFPT_Dielectric = "TPSS+U DFPT Dielectric"
    TPSS_U_NMR_Nuclear_Shielding = "TPSS+U NMR Nuclear Shielding"
    TPSS_U_NMR_Electric_Field_Gradient = "TPSS+U NMR Electric Field Gradient"
    TPSS_U_Static = "TPSS+U Static"
    TPSS_U_Structure_Optimization = "TPSS+U Structure Optimization"
    TPSS_U_Deformation = "TPSS+U Deformation"
    TPSS_U_Optic = "TPSS+U Optic"
    TPSS_U_Molecular_Dynamics = "TPSS+U Molecular Dynamics"
    TPSS_U_Unrecognized = "TPSS+U Unrecognized"
    R2SCAN_rVV10_U_NSCF_Line = "R2SCAN-rVV10+U NSCF Line"
    R2SCAN_rVV10_U_NSCF_Uniform = "R2SCAN-rVV10+U NSCF Uniform"
    R2SCAN_rVV10_U_Dielectric = "R2SCAN-rVV10+U Dielectric"
    R2SCAN_rVV10_U_DFPT = "R2SCAN-rVV10+U DFPT"
    R2SCAN_rVV10_U_DFPT_Dielectric = "R2SCAN-rVV10+U DFPT Dielectric"
    R2SCAN_rVV10_U_NMR_Nuclear_Shielding = "R2SCAN-rVV10+U NMR Nuclear Shielding"
    R2SCAN_rVV10_U_NMR_Electric_Field_Gradient = (
        "R2SCAN-rVV10+U NMR Electric Field Gradient"
    )
    R2SCAN_rVV10_U_Static = "R2SCAN-rVV10+U Static"
    R2SCAN_rVV10_U_Structure_Optimization = "R2SCAN-rVV10+U Structure Optimization"
    R2SCAN_rVV10_U_Deformation = "R2SCAN-rVV10+U Deformation"
    R2SCAN_rVV10_U_Optic = "R2SCAN-rVV10+U Optic"
    R2SCAN_rVV10_U_Molecular_Dynamics = "R2SCAN-rVV10+U Molecular Dynamics"
    R2SCAN_rVV10_U_Unrecognized = "R2SCAN-rVV10+U Unrecognized"
    SCAN_rVV10_U_NSCF_Line = "SCAN-rVV10+U NSCF Line"
    SCAN_rVV10_U_NSCF_Uniform = "SCAN-rVV10+U NSCF Uniform"
    SCAN_rVV10_U_Dielectric = "SCAN-rVV10+U Dielectric"
    SCAN_rVV10_U_DFPT = "SCAN-rVV10+U DFPT"
    SCAN_rVV10_U_DFPT_Dielectric = "SCAN-rVV10+U DFPT Dielectric"
    SCAN_rVV10_U_NMR_Nuclear_Shielding = "SCAN-rVV10+U NMR Nuclear Shielding"
    SCAN_rVV10_U_NMR_Electric_Field_Gradient = (
        "SCAN-rVV10+U NMR Electric Field Gradient"
    )
    SCAN_rVV10_U_Static = "SCAN-rVV10+U Static"
    SCAN_rVV10_U_Structure_Optimization = "SCAN-rVV10+U Structure Optimization"
    SCAN_rVV10_U_Deformation = "SCAN-rVV10+U Deformation"
    SCAN_rVV10_U_Optic = "SCAN-rVV10+U Optic"
    SCAN_rVV10_U_Molecular_Dynamics = "SCAN-rVV10+U Molecular Dynamics"
    SCAN_rVV10_U_Unrecognized = "SCAN-rVV10+U Unrecognized"
    optB86b_vdW_U_NSCF_Line = "optB86b-vdW+U NSCF Line"
    optB86b_vdW_U_NSCF_Uniform = "optB86b-vdW+U NSCF Uniform"
    optB86b_vdW_U_Dielectric = "optB86b-vdW+U Dielectric"
    optB86b_vdW_U_DFPT = "optB86b-vdW+U DFPT"
    optB86b_vdW_U_DFPT_Dielectric = "optB86b-vdW+U DFPT Dielectric"
    optB86b_vdW_U_NMR_Nuclear_Shielding = "optB86b-vdW+U NMR Nuclear Shielding"
    optB86b_vdW_U_NMR_Electric_Field_Gradient = (
        "optB86b-vdW+U NMR Electric Field Gradient"
    )
    optB86b_vdW_U_Static = "optB86b-vdW+U Static"
    optB86b_vdW_U_Structure_Optimization = "optB86b-vdW+U Structure Optimization"
    optB86b_vdW_U_Deformation = "optB86b-vdW+U Deformation"
    optB86b_vdW_U_Optic = "optB86b-vdW+U Optic"
    optB86b_vdW_U_Molecular_Dynamics = "optB86b-vdW+U Molecular Dynamics"
    optB86b_vdW_U_Unrecognized = "optB86b-vdW+U Unrecognized"
    optB88_vdW_U_NSCF_Line = "optB88-vdW+U NSCF Line"
    optB88_vdW_U_NSCF_Uniform = "optB88-vdW+U NSCF Uniform"
    optB88_vdW_U_Dielectric = "optB88-vdW+U Dielectric"
    optB88_vdW_U_DFPT = "optB88-vdW+U DFPT"
    optB88_vdW_U_DFPT_Dielectric = "optB88-vdW+U DFPT Dielectric"
    optB88_vdW_U_NMR_Nuclear_Shielding = "optB88-vdW+U NMR Nuclear Shielding"
    optB88_vdW_U_NMR_Electric_Field_Gradient = (
        "optB88-vdW+U NMR Electric Field Gradient"
    )
    optB88_vdW_U_Static = "optB88-vdW+U Static"
    optB88_vdW_U_Structure_Optimization = "optB88-vdW+U Structure Optimization"
    optB88_vdW_U_Deformation = "optB88-vdW+U Deformation"
    optB88_vdW_U_Optic = "optB88-vdW+U Optic"
    optB88_vdW_U_Molecular_Dynamics = "optB88-vdW+U Molecular Dynamics"
    optB88_vdW_U_Unrecognized = "optB88-vdW+U Unrecognized"
    optPBE_vdW_U_NSCF_Line = "optPBE-vdW+U NSCF Line"
    optPBE_vdW_U_NSCF_Uniform = "optPBE-vdW+U NSCF Uniform"
    optPBE_vdW_U_Dielectric = "optPBE-vdW+U Dielectric"
    optPBE_vdW_U_DFPT = "optPBE-vdW+U DFPT"
    optPBE_vdW_U_DFPT_Dielectric = "optPBE-vdW+U DFPT Dielectric"
    optPBE_vdW_U_NMR_Nuclear_Shielding = "optPBE-vdW+U NMR Nuclear Shielding"
    optPBE_vdW_U_NMR_Electric_Field_Gradient = (
        "optPBE-vdW+U NMR Electric Field Gradient"
    )
    optPBE_vdW_U_Static = "optPBE-vdW+U Static"
    optPBE_vdW_U_Structure_Optimization = "optPBE-vdW+U Structure Optimization"
    optPBE_vdW_U_Deformation = "optPBE-vdW+U Deformation"
    optPBE_vdW_U_Optic = "optPBE-vdW+U Optic"
    optPBE_vdW_U_Molecular_Dynamics = "optPBE-vdW+U Molecular Dynamics"
    optPBE_vdW_U_Unrecognized = "optPBE-vdW+U Unrecognized"
    rev_vdW_DF2_U_NSCF_Line = "rev-vdW-DF2+U NSCF Line"
    rev_vdW_DF2_U_NSCF_Uniform = "rev-vdW-DF2+U NSCF Uniform"
    rev_vdW_DF2_U_Dielectric = "rev-vdW-DF2+U Dielectric"
    rev_vdW_DF2_U_DFPT = "rev-vdW-DF2+U DFPT"
    rev_vdW_DF2_U_DFPT_Dielectric = "rev-vdW-DF2+U DFPT Dielectric"
    rev_vdW_DF2_U_NMR_Nuclear_Shielding = "rev-vdW-DF2+U NMR Nuclear Shielding"
    rev_vdW_DF2_U_NMR_Electric_Field_Gradient = (
        "rev-vdW-DF2+U NMR Electric Field Gradient"
    )
    rev_vdW_DF2_U_Static = "rev-vdW-DF2+U Static"
    rev_vdW_DF2_U_Structure_Optimization = "rev-vdW-DF2+U Structure Optimization"
    rev_vdW_DF2_U_Deformation = "rev-vdW-DF2+U Deformation"
    rev_vdW_DF2_U_Optic = "rev-vdW-DF2+U Optic"
    rev_vdW_DF2_U_Molecular_Dynamics = "rev-vdW-DF2+U Molecular Dynamics"
    rev_vdW_DF2_U_Unrecognized = "rev-vdW-DF2+U Unrecognized"
    revPBE_vdW_U_NSCF_Line = "revPBE-vdW+U NSCF Line"
    revPBE_vdW_U_NSCF_Uniform = "revPBE-vdW+U NSCF Uniform"
    revPBE_vdW_U_Dielectric = "revPBE-vdW+U Dielectric"
    revPBE_vdW_U_DFPT = "revPBE-vdW+U DFPT"
    revPBE_vdW_U_DFPT_Dielectric = "revPBE-vdW+U DFPT Dielectric"
    revPBE_vdW_U_NMR_Nuclear_Shielding = "revPBE-vdW+U NMR Nuclear Shielding"
    revPBE_vdW_U_NMR_Electric_Field_Gradient = (
        "revPBE-vdW+U NMR Electric Field Gradient"
    )
    revPBE_vdW_U_Static = "revPBE-vdW+U Static"
    revPBE_vdW_U_Structure_Optimization = "revPBE-vdW+U Structure Optimization"
    revPBE_vdW_U_Deformation = "revPBE-vdW+U Deformation"
    revPBE_vdW_U_Optic = "revPBE-vdW+U Optic"
    revPBE_vdW_U_Molecular_Dynamics = "revPBE-vdW+U Molecular Dynamics"
    revPBE_vdW_U_Unrecognized = "revPBE-vdW+U Unrecognized"
    vdW_DF2_U_NSCF_Line = "vdW-DF2+U NSCF Line"
    vdW_DF2_U_NSCF_Uniform = "vdW-DF2+U NSCF Uniform"
    vdW_DF2_U_Dielectric = "vdW-DF2+U Dielectric"
    vdW_DF2_U_DFPT = "vdW-DF2+U DFPT"
    vdW_DF2_U_DFPT_Dielectric = "vdW-DF2+U DFPT Dielectric"
    vdW_DF2_U_NMR_Nuclear_Shielding = "vdW-DF2+U NMR Nuclear Shielding"
    vdW_DF2_U_NMR_Electric_Field_Gradient = "vdW-DF2+U NMR Electric Field Gradient"
    vdW_DF2_U_Static = "vdW-DF2+U Static"
    vdW_DF2_U_Structure_Optimization = "vdW-DF2+U Structure Optimization"
    vdW_DF2_U_Deformation = "vdW-DF2+U Deformation"
    vdW_DF2_U_Optic = "vdW-DF2+U Optic"
    vdW_DF2_U_Molecular_Dynamics = "vdW-DF2+U Molecular Dynamics"
    vdW_DF2_U_Unrecognized = "vdW-DF2+U Unrecognized"
    LDA_NSCF_Line = "LDA NSCF Line"
    LDA_NSCF_Uniform = "LDA NSCF Uniform"
    LDA_Dielectric = "LDA Dielectric"
    LDA_DFPT = "LDA DFPT"
    LDA_DFPT_Dielectric = "LDA DFPT Dielectric"
    LDA_NMR_Nuclear_Shielding = "LDA NMR Nuclear Shielding"
    LDA_NMR_Electric_Field_Gradient = "LDA NMR Electric Field Gradient"
    LDA_Static = "LDA Static"
    LDA_Structure_Optimization = "LDA Structure Optimization"
    LDA_Deformation = "LDA Deformation"
    LDA_Optic = "LDA Optic"
    LDA_Molecular_Dynamics = "LDA Molecular Dynamics"
    LDA_Unrecognized = "LDA Unrecognized"
    LDA_U_NSCF_Line = "LDA+U NSCF Line"
    LDA_U_NSCF_Uniform = "LDA+U NSCF Uniform"
    LDA_U_Dielectric = "LDA+U Dielectric"
    LDA_U_DFPT = "LDA+U DFPT"
    LDA_U_DFPT_Dielectric = "LDA+U DFPT Dielectric"
    LDA_U_NMR_Nuclear_Shielding = "LDA+U NMR Nuclear Shielding"
    LDA_U_NMR_Electric_Field_Gradient = "LDA+U NMR Electric Field Gradient"
    LDA_U_Static = "LDA+U Static"
    LDA_U_Structure_Optimization = "LDA+U Structure Optimization"
    LDA_U_Deformation = "LDA+U Deformation"
    LDA_U_Optic = "LDA+U Optic"
    LDA_U_Molecular_Dynamics = "LDA+U Molecular Dynamics"
    LDA_U_Unrecognized = "LDA+U Unrecognized"
